@@ -11,6 +11,11 @@
 
 #undef register_class_attribute
 
+#ifdef HAVE_DMD_HELPER
+#  define WANT_DMD_API_044
+#  include "DMD_helper.h"
+#endif
+
 #include "perl-backcompat.c.inc"
 #include "sv_setrv.c.inc"
 
@@ -18,6 +23,7 @@
 #include "force_list_keeping_pushmark.c.inc"
 #include "optree-additions.c.inc"
 #include "newOP_CUSTOM.c.inc"
+#include "cv_copy_flags.c.inc"
 
 #ifdef DEBUGGING
 #  define DEBUG_OVERRIDE_PLCURCOP
@@ -27,23 +33,8 @@
 #  define DEBUG_SET_CURCOP_LINE(line)
 #endif
 
-#define need_PLparser()  S_need_PLparser(aTHX)
-static void S_need_PLparser(pTHX)
-{
-  if(!PL_parser) {
-    /* We need to generate just enough of a PL_parser to keep newSTATEOP()
-     * happy, otherwise it will SIGSEGV (RT133258)
-     */
-    SAVEVPTR(PL_parser);
-    Newxz(PL_parser, 1, yy_parser);
-    SAVEFREEPV(PL_parser);
-
-    PL_parser->copline = NOLINE;
-#if HAVE_PERL_VERSION(5, 20, 0)
-    PL_parser->preambling = NOLINE;
-#endif
-  }
-}
+#define need_PLparser()  ObjectPad__need_PLparser(aTHX)
+void ObjectPad__need_PLparser(pTHX); /* in Object/Pad.xs */
 
 /* Empty MGVTBL simply for locating instance backing AV */
 static MGVTBL vtbl_backingav = {};
@@ -149,13 +140,11 @@ void ObjectPad_mop_class_apply_attribute(pTHX_ ClassMeta *classmeta, const char
 
 /* TODO: get attribute */
 
-#define get_classmeta_for(self)  S_get_classmeta_for(aTHX_ self)
-static ClassMeta *S_get_classmeta_for(pTHX_ SV *self)
+ClassMeta *ObjectPad_mop_get_class_for_stash(pTHX_ HV *stash)
 {
-  HV *selfstash = SvSTASH(SvRV(self));
-  GV **gvp = (GV **)hv_fetchs(selfstash, "META", 0);
+  GV **gvp = (GV **)hv_fetchs(stash, "META", 0);
   if(!gvp)
-    croak("Unable to find ClassMeta for %" SVf, SVfARG(HvNAME(selfstash)));
+    croak("Unable to find ClassMeta for %" HEKf, HEKfARG(HvNAME_HEK(stash)));
 
   return NUM2PTR(ClassMeta *, SvUV(SvRV(GvSV(*gvp))));
 }
@@ -246,7 +235,7 @@ SV *ObjectPad_get_obj_backingav(pTHX_ SV *self, enum ReprType repr, bool create)
        *   https://rt.cpan.org/Ticket/Display.html?id=132263
        */
       if(!backingsvp) {
-        struct ClassMeta *classmeta = get_classmeta_for(self);
+        struct ClassMeta *classmeta = mop_get_class_for_stash(SvSTASH(rv));
         AV *backingav = newAV();
 
         make_instance_fields(classmeta, backingav, 0);
@@ -285,7 +274,10 @@ static CV *S_embed_cv(pTHX_ CV *cv, RoleEmbedding *embedding)
   assert(cv);
   assert(CvOUTSIDE(cv));
 
-  CV *embedded_cv = cv_clone(cv);
+  /* Perl core's cv_clone() would break in some situation here; see
+   *   https://rt.cpan.org/Ticket/Display.html?id=141483
+   */
+  CV *embedded_cv = cv_copy_flags(cv, 0);
   SV *embeddingsv = embedding->embeddingsv;
 
   assert(SvTYPE(embeddingsv) == SVt_PV && SvLEN(embeddingsv) >= sizeof(RoleEmbedding));
@@ -404,33 +396,16 @@ void ObjectPad_mop_class_add_BUILD(pTHX_ ClassMeta *meta, CV *cv)
 void ObjectPad_mop_class_add_ADJUST(pTHX_ ClassMeta *meta, CV *cv)
 {
   if(meta->sealed)
-    croak("Cannot add an ADJUST block to an already-sealed class");
+    croak("Cannot add an ADJUST(PARAMS) block to an already-sealed class");
   if(!meta->adjustblocks)
     meta->adjustblocks = newAV();
 
   AdjustBlock *block;
   Newx(block, 1, struct AdjustBlock);
 
-  block->is_adjustparams = false;
   block->cv = cv;
 
-  av_push(meta->adjustblocks, (SV *)block);
-}
-
-void ObjectPad_mop_class_add_ADJUSTPARAMS(pTHX_ ClassMeta *meta, CV *cv)
-{
-  if(meta->sealed)
-    croak("Cannot add an ADJUSTPARAMS block to an already-sealed class");
-  if(!meta->adjustblocks)
-    meta->adjustblocks = newAV();
-
-  AdjustBlock *block;
-  Newx(block, 1, struct AdjustBlock);
-
-  block->is_adjustparams = true;
-  block->cv = cv;
-
-  meta->has_adjustparams = true;
+  meta->has_adjust = true;
 
   av_push(meta->adjustblocks, (SV *)block);
 }
@@ -520,14 +495,11 @@ static RoleEmbedding *S_embed_role(pTHX_ ClassMeta *classmeta, ClassMeta *roleme
 
     CV *embedded_cv = embed_cv(block->cv, embedding);
 
-    if(block->is_adjustparams)
-      mop_class_add_ADJUSTPARAMS(classmeta, embedded_cv);
-    else
-      mop_class_add_ADJUST(classmeta, embedded_cv);
+    mop_class_add_ADJUST(classmeta, embedded_cv);
   }
 
-  if(rolemeta->has_adjustparams)
-    classmeta->has_adjustparams = true;
+  if(rolemeta->has_adjust)
+    classmeta->has_adjust = true;
 
   U32 nmethods = av_count(rolemeta->direct_methods);
   for(i = 0; i < nmethods; i++) {
@@ -542,6 +514,7 @@ static RoleEmbedding *S_embed_role(pTHX_ ClassMeta *classmeta, ClassMeta *roleme
     {
       MethodMeta *dstmethodmeta = mop_class_add_method(classmeta, mname);
       dstmethodmeta->role = rolemeta;
+      dstmethodmeta->is_common = methodmeta->is_common;
     }
 
     GV **gvp = (GV **)hv_fetch(dststash, SvPVX(mname), SvCUR(mname), GV_ADD);
@@ -1295,7 +1268,7 @@ XS_INTERNAL(injected_constructor)
   }
 
   HV *paramhv = NULL;
-  if(meta->parammap || meta->has_adjustparams || meta->strict_params) {
+  if(meta->parammap || meta->has_adjust || meta->strict_params) {
     paramhv = newHV();
     SAVEFREESV((SV *)paramhv);
 
@@ -1385,7 +1358,7 @@ XS_INTERNAL(injected_constructor)
 
       PUSHMARK(SP);
       PUSHs(self);
-      if(paramhv && block->is_adjustparams)
+      if(paramhv)
         mPUSHs(newRV_inc((SV *)paramhv));
       PUTBACK;
 
@@ -1522,7 +1495,7 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
   meta->sealed = false;
   meta->role_is_invokable = false;
   meta->strict_params = false;
-  meta->has_adjustparams = false;
+  meta->has_adjust = false;
   meta->has_superclass = false;
   meta->start_fieldix = 0;
   meta->next_fieldix = -1;
@@ -1714,8 +1687,8 @@ void ObjectPad_mop_class_set_superclass(pTHX_ ClassMeta *meta, SV *superclassnam
       }
     }
 
-    if(supermeta->has_adjustparams)
-      meta->has_adjustparams = true;
+    if(supermeta->has_adjust)
+      meta->has_adjust = true;
 
     U32 nroles;
     RoleEmbedding **embeddings = mop_class_get_all_roles(supermeta, &nroles);
@@ -1943,23 +1916,15 @@ static const struct ClassHookFuncs classhooks_strict = {
   .apply = &classhook_strict_apply,
 };
 
-void ObjectPad__boot_classes(void)
+void ObjectPad__boot_classes(pTHX)
 {
   register_class_attribute("isa",    &classhooks_isa,    NULL);
   register_class_attribute("does",   &classhooks_does,   NULL);
   register_class_attribute("repr",   &classhooks_repr,   NULL);
   register_class_attribute("compat", &classhooks_compat, NULL);
   register_class_attribute("strict", &classhooks_strict, NULL);
-}
 
-
-/* back-compat */
-SV *ObjectPad_get_obj_slotsav(pTHX_ SV *self, enum ReprType repr, bool create)
-{
-  return ObjectPad_get_obj_backingav(aTHX_ self, repr, create);
-}
-
-FieldMeta *ObjectPad_mop_class_add_slot(pTHX_ ClassMeta *meta, SV *fieldname)
-{
-  return ObjectPad_mop_class_add_field(aTHX_ meta, fieldname);
+#ifdef HAVE_DMD_HELPER
+  DMD_ADD_ROOT((SV *)&vtbl_backingav, "the Object::Pad backing AV VTBL");
+#endif
 }
